/*
Copyright (c) 2021 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#undef _WIN32_WINNT
#define _WIN32_WINNT 0x0A00
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#define BUFMAX 194 // maximum number of INPUT_RECORD structures that we get on the stack
#define HELP                                                                        \
  "Print to a Virtual Terminal emulator, read response sequences.\r\n"              \
  "vt.exe out\r\n"                                                                  \
  "  out  String of characters that may contain VT escape sequences.\r\n\r\n"       \
  "Use \"\\n\" to write a new line.\r\n"                                            \
  "It is possible to use \"\\e\" rather than an ESC character in VT sequences.\r\n" \
  "(Double the backslashes for literal \"\\n\" or \"\\e\" output.)\r\n"             \
  "Responses of the VT emulator are read and printed to stdout.\r\n\r\n"

#ifdef _MSC_VER
# pragma warning(push)
# pragma warning(disable : 4706 6001)
#endif

// skip the executable name and return a pointer to the program arguments in the command line
static wchar_t *args(wchar_t *cmdLn)
{
  for (; *cmdLn != L' ' && *cmdLn != L'\t' && *cmdLn; ++cmdLn)
    if (*cmdLn == L'"')
      for (++cmdLn; *cmdLn != L'"' && *cmdLn; ++cmdLn);

  for (int cnt = 0; cnt < 2 && (*cmdLn == L' ' || *cmdLn == L'\t') && *cmdLn; ++cnt, ++cmdLn);
  return cmdLn;
}

// replace "\e" and "\n" in the arguments with ESC and CR LF characters respectively
static DWORD unesc(wchar_t *const str)
{
  wchar_t *outIter = str;
  for (const wchar_t *inIter = str; *inIter; ++inIter, ++outIter)
  {
    switch (*inIter)
    {
      case L'\\':
        {
          const wchar_t *const p2nd = inIter + 1;
          if (*p2nd == L'E' || *p2nd == L'e')
          {
            *outIter = L'\033';
            inIter = p2nd;
            break;
          }
          if (*p2nd == L'N' || *p2nd == L'n')
          {
            *outIter = L'\r';
            *++outIter = L'\n';
            inIter = p2nd;
            break;
          }
          if (*p2nd == L'\\')
          {
            const wchar_t *const p3rd = inIter + 2;
            if (*p3rd == L'E' || *p3rd == L'N' || *p3rd == L'e' || *p3rd == L'n')
            {
              *outIter = L'\\';
              *++outIter = *p3rd;
              inIter = p3rd;
              break;
            }
          }
        }
      // fall through
      default:
        *outIter = *inIter;
    }
  }

  return (DWORD)(outIter - str);
}

// Defining this macro will only affect the GCC compiler and can reduce the size of the utility. As soon as you define it you have to add these linker options:
//   -Wl,-e,__start -nostartfiles
#define USE_START

#if defined(__GNUC__) && defined(USE_START)
# include <stdlib.h>
# define QUIT(_status) exit(_status)
# ifdef __i386
void _start(void);
void _start(void)
# else
void __start(void);
void __start(void)
# endif
#else
# define QUIT(_status) return (_status)
int main(void)
#endif
{
  const HANDLE hStdOut = GetStdHandle(STD_OUTPUT_HANDLE); // we expect this handle to stdout to be redirected in order to read response sequences
  if (hStdOut == INVALID_HANDLE_VALUE)
    QUIT(1);

  DWORD length = 0;
  wchar_t *argStr = args(GetCommandLineW());
  if (*argStr == L'/' && argStr[1] == L'?' && argStr[2] == L'\0') // argument is "/?"
  {
    WriteFile(hStdOut, HELP, (DWORD)(sizeof(HELP) - 1), &length, NULL); // print the help message
    QUIT(0);
  }

  const HANDLE hConOut = CreateFileA("CONOUT$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL);
  if (hConOut == INVALID_HANDLE_VALUE)
    QUIT(1);

  DWORD mode;
  BOOL isGood =
    GetConsoleMode(hConOut, &mode) && // capture the current mode of the console output device
    (SetConsoleMode(hConOut, mode | ENABLE_VIRTUAL_TERMINAL_PROCESSING | DISABLE_NEWLINE_AUTO_RETURN) || SetConsoleMode(hConOut, mode | ENABLE_VIRTUAL_TERMINAL_PROCESSING)); // ensure VT processing is enabled (this will fail if VT processing is not supported)

  if (!isGood || !*argStr)
  {
    SetConsoleMode(hConOut, mode);
    CloseHandle(hConOut);
    QUIT(!isGood); // this is intended to tell the user if VT processing is supported
  }

  HANDLE hConIn;
  isGood =
    (hConIn = CreateFileA("CONIN$", GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE, NULL, OPEN_EXISTING, 0, NULL)) != INVALID_HANDLE_VALUE && // we need to read the response from the console input device and thus, open a handle to it
    FlushConsoleInputBuffer(hConIn) && // discard all remaining input
    (length = unesc(argStr)) && // turn "\e" into an ESC character, and "\n" into CR and LF characters (case-sensitive each)
    WriteConsoleW(hConOut, argStr, length, &length, NULL) && // write the passed arguments
    SetConsoleMode(hConOut, mode); // restore the original mode of the console output device

  CloseHandle(hConOut);
  INPUT_RECORD inBuf[BUFMAX];
  if (isGood) // we may get a response from the VT emulator
  {
    length = 0;
    PeekConsoleInputA(hConIn, inBuf, BUFMAX, &length); // fetch pending input events (Peek... in order to avoid blocking if we did not receive a response)
    FlushConsoleInputBuffer(hConIn); // Peek... did not remove the read input from the queue, we have to flush it explicitly
  }

  if (hConIn != INVALID_HANDLE_VALUE)
    CloseHandle(hConIn);

  if (!isGood) // something went wrong, we get outta here
    QUIT(1);

  if (!length) // no input events captured
    QUIT(0);

  char *outBuf = (char*)inBuf, *outIter = outBuf; // checked it, no race condition if we re-use the input buffer for the output
  for (const INPUT_RECORD *inIter = inBuf, *const inEnd = inBuf + length; inIter < inEnd; ++inIter) // iterate over the captured input events
    if (inIter->EventType == KEY_EVENT && inIter->Event.KeyEvent.bKeyDown) // take only key-down events
      *outIter++ = inIter->Event.KeyEvent.uChar.AsciiChar; // cache the character belonging to the event

  if (!(length = (DWORD)(outIter - outBuf))) // no characters in our cache
    QUIT(0);

  SetConsoleMode(hStdOut, mode | ENABLE_VIRTUAL_TERMINAL_PROCESSING); // we want the VT processor to discard the response sequence if stdout is not redirected
  WriteFile(hStdOut, outBuf, length, &length, NULL); // write the cache to stdout
  SetConsoleMode(hStdOut, mode);
  QUIT(0);
}

#ifdef _MSC_VER
# pragma warning(pop)
#endif
